from inspect import getmembers, isfunction, ismethod
import json
import os
import re
import time

import boto3
import botocore
from botocore.exceptions import ClientError

from trailblazer import log
from trailblazer.boto.util import botocore_config


def get_service_json_files(config):
    service_file = {}

    root_dir = config['botocore_document_json_path']

    for service_dir in os.listdir(root_dir):
        if os.path.isdir(os.path.join(root_dir,service_dir)):
            date_dirs = os.listdir(os.path.join(root_dir,service_dir))
            date_dir = None
            if len(date_dirs) > 1:
                date_dir = date_dirs[-1]
            else:
                date_dir = date_dirs[0]
            if os.path.exists(os.path.join(root_dir, service_dir, date_dir, 'service-2.json')):
                service_file[service_dir] = os.path.join(root_dir, service_dir, date_dir, 'service-2.json')
            else:
                service_file[service_dir] = None

    return service_file

def get_service_call_params(service_json_file):

    services = {}

    json_data = json.load(open(service_json_file))

    for key, value in json_data.get('operations', {}).items():
        services[key.lower()] = value.get('http', {}).get('requestUri', '/')

    return services


def get_service_call_mutation(service_json_file):

    services = {}

    json_data = json.load(open(service_json_file))

    for key, value in json_data.get('operations', {}).items():

        method = value.get('http', {}).get('method', 'UNKOWN')

        if method == 'GET':
            services[key.lower()] = 'nonmutating'
        else:
            services[key.lower()] = 'mutating'

    return services


def get_boto_functions(client):
    """Loop through the client member functions and pull out what we can actually call"""
    functions_list = [o for o in getmembers(client) if ( ( isfunction(o[1]) or ismethod(o[1]) )
            and not o[0].startswith('_') and not o[0] == 'can_paginate' and not o[0] == 'get_paginator'
            and not o[0] == 'get_waiter' and 'presigned' not in o[0])]

    return functions_list


def make_api_call(service, function, region, func_params):

    if function[0] == 'generate_presigned_url':
        func_params['ClientMethod'] = 'get_object'

    if service == 's3':
        if function[0] == 'copy':
            copy_source = {
                'Bucket': 'mybucket',
                'Key': 'mykey'
            }
            function[1](copy_source, 'testbucket', 'testkey')
            time.sleep(.1)
            return
        elif function[0] == 'download_file':
            function[1]('testbucket', 'testkey', 'test')
            time.sleep(.1)
            return
        elif function[0] == 'download_fileobj':
            with open('testfile', 'wb') as data:
                function[1]('testbucket', 'testkey', data)
                time.sleep(.1)
                return
        elif function[0] == 'upload_file':
            function[1](service_file_json[service], 'testbucket', 'testkey')
            time.sleep(.1)
            return
        elif function[0] == 'upload_fileobj':
            with open(service_file_json[service], 'rb') as data:
                function[1](data, 'testbucket', 'testkey')
                time.sleep(.1)
                return
    elif service == 'ec2':
        if function[0] == 'copy_snapshot':
            function[1](SourceRegion=region, **func_params)
            time.sleep(.1)

    function[1](**func_params)
    time.sleep(.1)
    return



import boto3
from botocore.exceptions import ClientError

from trailblazer import log


def get_assume_role_session(account_number, role, session_id):
    arn = "arn:aws:iam::{}:role/{}".format(account_number, role)

    try:
    	session = boto3.Session()
    	sts = session.client('sts')
    	assumed_role = sts.assume_role(RoleArn=arn, RoleSessionName=session_id)
    except ClientError as e:
    	log.fatal(e)
    else:
    	session = boto3.Session(
    		aws_access_key_id=assumed_role['Credentials']['AccessKeyId'],
    		aws_secret_access_key=assumed_role['Credentials']['SecretAccessKey'],
    		aws_session_token=assumed_role['Credentials']['SessionToken']
    	)
    	return session



import re

import boto3
from botocore.exceptions import ClientError

from trailblazer import log
from trailblazer.boto.util import botocore_config
from trailblazer.boto.service import get_boto_functions, get_service_call_params, \
									get_service_json_files, make_api_call, get_service_call_mutation
from trailblazer.boto.sts import get_assume_role_session


def enumerate_services(config, services, dry_run=False):

    # Create a boto3 session to use for enumeration
    session = boto3.Session()

    authorized_calls = []

    for service in services:

        if len(session.get_available_regions(service)) == 0:
            log.debug('Skipping {} - No regions exist for this service'.format(service))
            continue

        # Create a service client
        log.info('Creating {} client...'.format(service))

        # Grab a region to use for the calls.  This should be us-west-2
        region = session.get_available_regions(service)[-1]

        # Set the user-agent if specified in the config
        if config.get('user_agent', None):
            botocore_config.user_agent = config['user_agent']

        # Create a client with parameter validation off
        client = session.client(service, region_name=region, config=botocore_config)

        # Get the functions that you can call
        functions_list = get_boto_functions(client)

        # Get the service file
        service_file_json = get_service_json_files(config)

        # Get a list of params needed to make the serialization pass in botocore
        service_call_params = get_service_call_params(service_file_json[service])

        # Loop through all the functions and call them
        for function in functions_list:

            # The service_file_json doesn't have underscores in names so let's remove them
            function_key = function[0].replace('_','')

            # Session Name Can only be 64 characters long
            if len(function_key) > 64:
                session_name = function_key[:63]
                log.info('Session Name {} is for {}'.format(session_name, function_key))
            else:
                session_name = function_key

            # Set the session to the name of the API call we are making
            session = get_assume_role_session(
                account_number=config['account_number'],
                role=config['account_role'],
                session_id=session_name
            )

            new_client = session.client(service, region_name=region, config=botocore_config)
            new_functions_list = get_boto_functions(new_client)

            for new_func in new_functions_list:
                if new_func[0] == function[0]:

                    # We need to pull out the parameters needed in the requestUri, ex. /{Bucket}/{Key+} -> ['Bucket', 'Key']
                    params = re.findall('\{(.*?)\}', service_call_params.get(function_key, '/'))
                    params = [p.strip('+') for p in params]

                    try:
                        func_params = {}

                        for param in params:
                            # Set something because we have to
                            func_params[param] = 'testparameter'

                        log.info('Calling {}.{} with params {} in {}'.format(service, new_func[0], func_params, region))

                        if not dry_run:
                        	make_api_call(service, new_func, region, func_params)

                    except ClientError as e:
                        log.debug(e)
                    except boto3.exceptions.S3UploadFailedError as e:
                        log.debug(e)
                    except TypeError as e:
                        log.debug(e)
                    except KeyError as e:
                        log.debug('Unknown Exception: {}.{} - {}'.format(service, new_func[0], e))



from __future__ import absolute_import, division, print_function

__all__ = [
    '__title__', '__summary__', '__uri__', '__version__', '__author__',
    '__email__', '__license__', '__copyright__',
]

__title__ = 'trailblazer-aws'
__summary__ = ('Blazing Cloudtrail since 2018')
__uri__ = 'https://github.com/willbengtson/trailblazer-aws'

__version__ = '1.0.1'

__author__ = 'Will Bengtson'
__email__ = 'william.bengtson@gmail.com'

__license__ = 'Apache License, Version 2.0'
__copyright__ = 'Copyright 2018 {0}'.format(__author__)
\


import logging
import json
from inspect import getmembers, isfunction, ismethod
import os
import sys
import yaml

import boto3
import click
import click_log
from click.exceptions import UsageError

from trailblazer import log
from trailblazer.__about__ import __version__
from trailblazer.boto.service import get_service_json_files, get_service_call_params
from trailblazer.attack import simulate_attack
from trailblazer.cloudtrail import process_cloudtrail, record_cloudtrail
from trailblazer.enumerate import enumerate_services


click_log.basic_config(log)


class YAML(click.ParamType):
    name = 'yaml'

    def convert(self, value, param, ctx):
        try:
            with open(value, 'rb') as f:
                return yaml.safe_load(f.read())
        except (IOError, OSError) as e:
            self.fail('Could not open file: {0}'.format(value))


class CommaList(click.ParamType):
    name = 'commalist'

    def convert(self, value, param, ctx):
        return value.split(',')


class AppContext(object):
    def __init__(self):
        self.config = None
        self.dry_run = False


pass_context = click.make_pass_decorator(AppContext, ensure=True)


@click.group()
@click_log.simple_verbosity_option(log)
@click.option('--config', type=YAML(), help='Configuration file to use.')
@click.option('--dry-run', type=bool, default=False, is_flag=True, help='Run command without persisting anything.')
@click.version_option(version=__version__)
@pass_context
def cli(ctx, config, dry_run):

    if not ctx.config:
        ctx.config = config

    if not ctx.dry_run:
        ctx.dry_run = dry_run

        log.debug('Current context. DryRun: {} Config: {}'.format(ctx.dry_run, json.dumps(ctx.config, indent=2)))

    if not ctx.config.get('botocore_document_json_path', None):
        log.fatal('botocore_document_json_path not defined in config file')


@cli.group()
def enumerate():
    pass


@cli.group()
def simulate():
    pass


@cli.group()
def cloudtrail():
    pass


@enumerate.command()
@click.option('--services', type=CommaList(), help='Comma delimited list of services')
@pass_context
def cloudtrail_calls(ctx, services):
    """Enumerate all calls for AWS Services to determine what shows up in CloudTrail"""
    log.info('Starting enumeration for CloudTrail...')

    session = boto3.Session()

    if not services:
        services = session.get_available_services()

    enumerate_services(ctx.config, services, dry_run=ctx.dry_run)

    log.info('Enumeration complete')


@simulate.command()
@pass_context
def attack(ctx):
    """Simulate an attack by making the calls described in the config"""
    if not ctx.config.get('attack_chain', None):
        log.fatal('attack_chain not found in config file')

    log.info('Starting attack simulation...')

    attack_commands = ctx.config['attack_chain']

    simulate_attack(ctx.config, attack_commands, dry_run=ctx.dry_run)

    log.info('Attack simulation complete')


@cloudtrail.command()
@click.option('--directory', type=str, help='Path to directory with CloudTrail files', required=True)
@click.option('--arn', type=str, help='User ARN making calls', required=True)
@pass_context
def process(ctx, directory, arn):
    """Process cloudtrail files"""
    log.info('Processing CloudTrail...')

    if not os.path.exists(directory):
        log.fatal('Invalid Directory Path')

    files = []
    for cloudtrail_file in os.listdir(directory):
        files.append(os.path.join(directory, cloudtrail_file))

    api_calls_logged = process_cloudtrail(arn, files)


@cloudtrail.command()
@click.option('--directory', type=str, help='Path to directory with CloudTrail files', required=True)
@click.option('--arn', type=str, help='User ARN making calls', required=True)
@click.option('--output', type=str, help='Output File Name')
@pass_context
def record(ctx, directory, arn, output):
    """Create attack simulation from CloudTrail"""
    log.info('Recording CloudTrail...')

    if not os.path.exists(directory):
        log.fatal('Invalid Directory Path')

    files = []
    for cloudtrail_file in os.listdir(directory):
        files.append(os.path.join(directory, cloudtrail_file))

    api_calls_recorded = record_cloudtrail(arn, files)

    if output:
        with open(output, 'w') as yaml_file:
            yaml.dump(
                {
                    'attack_chain': api_calls_recorded
                },
                yaml_file,
                default_flow_style=False
            )

if __name__ == '__main__':
    try:
        cli()
    except KeyboardInterrupt:
        logging.debug("Exiting due to KeyboardInterrupt...")




from datetime import datetime, timedelta
import gzip
import json

from trailblazer import log


def process_cloudtrail(arn, files):

    api_calls = []

    log.info('EventSource, EventName, Recorded Name, Match')

    for file in files:
        f = None
        log.debug('Processing file: {}'.format(file))

        if file.endswith('.gz'):
            f = gzip.open(file, 'r')
        else:
            f = open(file, 'r')

        try:
            cloudtrail = json.load(f)
        except Exception as e:
            log.error('Invalid JSON File: {} - {}'.format(file, e))
            continue

        for record in cloudtrail['Records']:
            if record.get('userIdentity', {}).get('arn', '').startswith(arn):
                event_source = record['eventSource'].split('.')[0]
                event_name = record['eventName']

                call = '{}.{}'.format(event_source, event_name)

                if call not in api_calls:
                    session = record['userIdentity']['arn'].split('/')[-1]
                    match = (record['eventName'].lower() == session)
                    log.info('{}, {}, {}, {}'.format(record['eventSource'].split('.')[0], record['eventName'], session, match))
                    api_calls.append(call)

        f.close()

    return api_calls


def pairwise(lst):
    """ yield item i and item i+1 in lst. e.g.
        (lst[0], lst[1]), (lst[1], lst[2]), ..., (lst[-1], None)
    """
    if not lst: return
    #yield None, lst[0]
    for i in range(len(lst)-1):
        yield lst[i], lst[i+1]
    yield lst[-1], None


def record_cloudtrail(arn, files):

    api_calls = []

    for file in files:
        f = None
        log.info('Processing file: {}'.format(file))

        if file.endswith('.gz'):
            f = gzip.open(file, 'r')
        else:
            f = open(file, 'r')

        try:
            cloudtrail = json.load(f)
        except Exception as e:
            log.error('Invalid JSON File: {} - {}'.format(file, e))
            continue

        records = sorted(cloudtrail['Records'], key=lambda x: datetime.strptime(x['eventTime'], '%Y-%m-%dT%H:%M:%SZ'), reverse=False)

        for record, next_record in pairwise(records):
            if record.get('userIdentity', {}).get('arn', '').startswith(arn):
                event_source = record['eventSource'].split('.')[0]
                event_name = record['eventName']

                time_delay = 0
                if next_record:
                    time_delta = datetime.strptime(next_record['eventTime'], '%Y-%m-%dT%H:%M:%SZ') - datetime.strptime(record['eventTime'], '%Y-%m-%dT%H:%M:%SZ')
                    time_delay = time_delta.seconds

                call = '{}.{}'.format(event_source, event_name)

                log.info('{}.{} - {}'.format(record['eventSource'].split('.')[0], record['eventName'], record['userIdentity']['arn']))

                api_calls.append(
                    {
                        'call': '{}.{}'.format(event_source, event_name),
                        'time_delay': time_delay
                    }
                )

        f.close()

    return api_calls



import json
import re
import time

import boto3

from trailblazer import log
from trailblazer.boto.service import get_boto_functions, get_service_json_files, get_service_call_params, make_api_call
from trailblazer.boto.util import botocore_config


def make_call(config, session, service, command, region=None):
    # Grab a region to use for the calls.  This should be us-west-2
    if not region:
        region = session.get_available_regions(service)[-1]

    if config.get('user_agent', None):
        botocore_config.user_agent = config['user_agent']

    # Create a client with parameter validation off
    client = session.client(service, region_name=region, config=botocore_config)

    # Get the functions that you can call
    functions_list = get_boto_functions(client)

    # Get the service file
    service_file_json = get_service_json_files(config)

    # Get a list of params needed to make the serialization pass in botocore
    service_call_params = get_service_call_params(service_file_json[service])


    for function in functions_list:
        if function[0] == command:

            # The service_file_json doesn't have underscores in names so let's remove them
            function_key = function[0].replace('_','')

            # We need to pull out the parameters needed in the requestUri, ex. /{Bucket}/{Key+} -> ['Bucket', 'Key']
            params = re.findall('\{(.*?)\}', service_call_params.get(function_key, '/'))
            params = [p.strip('+') for p in params]

            func_params = {}

            for param in params:
                func_params[param] = 'testparameter'

            try:
                make_api_call(service, function, region, func_params)
            except ClientError as e:
                log.debug(e)
            except boto3.exceptions.S3UploadFailedError as e:
                log.debug(e)
            except TypeError as e:
                log.debug(e)
            except KeyError as e:
                log.debug('Unknown Exception: {}.{} - {}'.format(service, function[0], e))

            return


def simulate_attack(config, commands, dry_run=False):
    log.debug('Attack chain to be executed:')
    log.debug(json.dumps(commands, indent=4))

    session = boto3.Session()

    for command in commands:
        service_event = command['call'].split('.')

        service = service_event[0]
        api_call = service_event[1]

        delay = command.get('time_delay', 0)
        region = command.get('region', None)

        log.info('Making call - {}.{}'.format(service, api_call))
        if not dry_run:
            make_call(config, session, service, api_call, region)
        log.info('Sleeping {} until next call'.format(delay))
        if not dry_run:
            time.sleep(delay)




import botocore

from trailblazer.__about__ import __version__


# Create custom botocore session with a custom config
botocore_config = botocore.config.Config(
    parameter_validation = False,
    user_agent = 'Trailblazer/{}'.format(__version__)
)



#     Copyright 2018 Will Bengtson
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
import re
import ast
import os.path
import sys
from setuptools import setup, find_packages

_version_re = re.compile(r'__version__\s+=\s+(.*)')
with open('trailblazer/__about__.py', 'rb') as f:
    TRAILBLAZER_VERSION = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

# When executing the setup.py, we need to be able to import ourselves.  This
# means that we need to add the src/ directory to the sys.path

sys.path.insert(0, ROOT)

install_requirements = [
    'boto3>=1.5.34',
    'click==6.7',
    'click-log==0.2.1',
    'PyYAML==3.12'
]

setup(
    name='trailblazer-aws',
    version=TRAILBLAZER_VERSION,
    long_description="CloudTrail enumeration and AWS attack platform",
    packages=find_packages(),
    install_requires=install_requirements,
    entry_points={
        'console_scripts': [
            'trailblazer = trailblazer.cli:cli',
        ],
    }
)